/** @file mission_node.cpp
*  @version 4.0
*  @date June 2020
*
*  @brief node of hotpoint 1.0/waypoint 1.0.
*
*  @Copyright (c) 2020 DJI
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
*/
#include <dji_osdk_ros/bus_driver.h>
#include <dji_osdk_ros/FlightTaskControl.h>
#include <dji_osdk_ros/ObtainControlAuthority.h>
#include <dji_osdk_ros/Overwatch.h>
#include <geometry_msgs/Transform.h>
#include <geometry_msgs/Vector3.h>
#include <geometry_msgs/Quaternion.h>
#include <tf2/LinearMath/Quaternion.h>
#include <string>
#include <iostream>
#include <fstream>

using namespace DJI::OSDK;
using namespace DJI::OSDK::Telemetry;

// global variables
ros::ServiceClient     waypoint_upload_client;
ros::ServiceClient     waypoint_action_client;
ros::ServiceClient     flight_control_client;
ros::ServiceClient     mission_waypoint_client;
ros::ServiceClient     obtain_ctrl_authority_client;
ros::Publisher         syncPub; // thing to publish to sync topic "positionSync"

sensor_msgs::NavSatFix gps_pos; // gps position from drone
geometry_msgs::Transform ex_pos; // position from external source (qualisys or some such)
bool                   gps_home_flag = false;
bool                   pos_home_flag = false;
bool                   positionMissionFlag = false;
bool                   waypointMissionFlag = false;
ros::Subscriber        gps_pos_subscriber;
ros::Subscriber        pos_feedback_subscriber;
int                    droneId; // stupid but will work for now

int                    waypointMissionUpdateCounter = 0; // counts iterations before we check waypoints again
int                    waypointMissionUpdateLimit = 10; // number of counts before we check ...
int                    positionMissionUpdateCounter = 0; // counts iterations before we check positions again
int                    positionMissionUpdateLimit = 10; // ...

std::vector<WayPointSettings> generatedWaypts; // waypoints to be generated, used to check if we're at waypoint or not
int                    waypointNumber;
int                    currentWaypoint = 0; // how many waypoints have we reached
long double            initialWaypointDistance; // distance to first waypoint
long double            initialPositionDistance; // distance to first position
double                 baseSpeedParam;
double                 minSpeedParam;
double                 speedScalarParam;

std::vector<std::vector<float>> poseRotationMatrix; // need to run calibration to get
float poseMagnitudeScaler; // need to run calibration to get
float poseRotationRadians;
bool rotationMatrixCalibratedFlag = false;

// Utility function for
// converting degrees to radians
long double toRadians(const long double degree)
{
  // cmath library in C++
  // defines the constant
  // M_PI as the value of
  // pi accurate to 1e-30
  long double one_deg = (M_PI) / 180;
  return (one_deg * degree);
}

float eucDistance(std::vector<float> a, std::vector<float> b)
{
  // do the math bb
  return pow(pow(a[0] - b[0],2) + pow(a[1] - b[1],2) + pow(a[2] - b[2],2),0.5);
}

long double gpsDistance(long double lat1, long double long1, long double lat2, long double long2)
{
  // Convert the latitudes
  // and longitudes
  // from degree to radians.
  lat1 = toRadians(lat1);
  long1 = toRadians(long1);
  lat2 = toRadians(lat2);
  long2 = toRadians(long2);

  // Haversine Formula
  long double dlong = long2 - long1;
  long double dlat = lat2 - lat1;

  long double ans = pow(sin(dlat / 2), 2) +
  cos(lat1) * cos(lat2) *
  pow(sin(dlong / 2), 2);

  ans = 2 * asin(sqrt(ans));

  // Radius of Earth in
  // Kilometers, R = 6371
  // Use R = 3956 for miles
  long double R = 6371;

  // Calculate the result
  ans = ans * R;

  // recalculate in meters
  ans = ans * 1000;

  return ans;
}

std::vector<std::vector<float>> rotateAxisAngle(std::vector<float> u, float angleRad)
{
  const float sinA = sinf( angleRad );
  const float cosA = cosf( angleRad );
  const float oneMinusCosA = 1.0f - cosA;

  // rotationMatrix
  std::vector<std::vector<float>> rotationMatrix
  {
    {(u[0] * u[0] * oneMinusCosA) + cosA, (u[1] * u[0] * oneMinusCosA) - (sinA * u[2]), (u[2] * u[0] * oneMinusCosA) + (sinA * u[1])},
    {(u[0] * u[1] * oneMinusCosA) + (sinA * u[2]), (u[1] * u[1] * oneMinusCosA) + cosA, (u[2] * u[1] * oneMinusCosA) - (sinA * u[0])},
    {(u[0] * u[2] * oneMinusCosA) - (sinA * u[1]), (u[1] * u[2] * oneMinusCosA) + (sinA * u[0]), (u[2] * u[2] * oneMinusCosA) + cosA}
  };
  return rotationMatrix;
}

// Function that return
// dot product of two vector array.
float dotProduct(std::vector<float> vect_A, std::vector<float> vect_B)
{

    float product = 0;
    int vectorSize = vect_A.size(); // both vectors must be of same length
    // Loop for calculate cot product
    for (int i = 0; i < vectorSize; i++)

        product = product + vect_A[i] * vect_B[i];
    return product;
}

// Function to find
// cross product of two vector array.
std::vector<float> crossProduct(std::vector<float> vect_A, std::vector<float> vect_B)
{
  std::vector<float> crossP
  {
    vect_A[1] * vect_B[2] - vect_A[2] * vect_B[1],
    vect_A[2] * vect_B[0] - vect_A[0] * vect_B[2],
    vect_A[0] * vect_B[1] - vect_A[1] * vect_B[0]
  };
  return crossP;
}

std::vector<float> normalizeVector(std::vector<float> vec)
{
  float mag = Magnitude3(vec);
  std::vector<float> result {
    vec[0]/mag,
    vec[1]/mag,
    vec[2]/mag
  };
  return result;
}

float Magnitude3(std::vector<float> v)
{
  return pow(pow(v[0], 2) + pow(v[1], 2) + pow(v[2], 2),0.5);
}

std::vector<float> MatrixMultiply(std::vector <std::vector<float>> a, std::vector<float> b)
{
    const int n = a.size();     // a cols
    const int m = a[0].size();  // a rows
    std::vector <float> c(n);

    // the PAIN
    // ROS_INFO("  rotation matrix calc:");

    for (auto k = 0; k < m; ++k) // row loop
    {
      for (auto i = 0; i < n; ++i) // col loop
      {
        // ROS_INFO("matrix (%d, %d): %f", k, i, a[i][k]);
        c[k] += a[i][k] * b[i];
      }
      // ROS_INFO("result val: %f", c[k]);
    }
    return c;
}

std::vector<std::vector<float>> rotateAlign(std::vector<float> v1, std::vector<float> v2)
{
    std::vector<float> axis = normalizeVector(crossProduct(v1, v2));
    float dotProd = dotProduct(normalizeVector(v1), normalizeVector(v2));
    if (dotProd < -1.0)
    {
      dotProd = -1.0;
    }
    else if (dotProd > 1.0)
    {
      dotProd = 1.0;
    }
    poseRotationRadians = acosf(dotProd);
    std::vector<std::vector<float>> result = rotateAxisAngle(axis, poseRotationRadians); // returns rotation matrix

    ROS_INFO("axis of rotation: (%f, %f, %f)", axis[0], axis[1], axis[2]);
    ROS_INFO("dot product: (%f)", dotProd);
    ROS_INFO("angle rad: (%f)", poseRotationRadians);
    return result;
}

void gpsPosCallback(const sensor_msgs::NavSatFix::ConstPtr& msg)
{
  gps_pos = *msg;

  if (!waypointMissionFlag)
  {
    return; // peace out if not running mission
  }

  if (!gps_home_flag)
  {
    int generatedWaypointNumber = static_cast<int>(generatedWaypts.size());
    if (generatedWaypointNumber > 0)
    {
      initialWaypointDistance = gpsDistance(gps_pos.latitude, gps_pos.longitude, generatedWaypts[0].latitude, generatedWaypts[0].longitude);
      ROS_INFO("current position: (%f, %f)", static_cast<float32_t>(gps_pos.latitude), static_cast<float32_t>(gps_pos.longitude));
      ROS_INFO("first waypoint position: (%f, %f)", static_cast<float32_t>(generatedWaypts[0].latitude), static_cast<float32_t>(generatedWaypts[0].longitude));

      ROS_INFO("distance to first waypoint: %f", static_cast<float32_t>(initialWaypointDistance));
      gps_home_flag = true;
    }
  }
  else
  {
    // do checking to see if we're near a waypoint, pause if so
    if(waypointMissionUpdateCounter >= waypointMissionUpdateLimit)
    {

      waypointMissionUpdateCounter = 0;
      int generatedWaypointNumber = static_cast<int>(generatedWaypts.size());
      if(generatedWaypointNumber > currentWaypoint) // if we actually have waypoints left
      {
        long double tolerance = 0.5; // reached waypoint detection threshold
        long double nextWaypointDistance;
        if (currentWaypoint == 0)
        {
          nextWaypointDistance = initialWaypointDistance;
        }
        else
        {
          nextWaypointDistance =
          gpsDistance(generatedWaypts[currentWaypoint-1].latitude, generatedWaypts[currentWaypoint-1].longitude, generatedWaypts[currentWaypoint].latitude, generatedWaypts[currentWaypoint].longitude);
        }

        long double progressDistance =
        gpsDistance(gps_pos.latitude, gps_pos.longitude, generatedWaypts[currentWaypoint].latitude, generatedWaypts[currentWaypoint].longitude);

        // now update sync topic with progress through waypoints
        std_msgs::Float32 syncUpdate;
        syncUpdate.data = static_cast<float32_t>(currentWaypoint) + static_cast<float32_t>(static_cast<float32_t>(1) - progressDistance/nextWaypointDistance);
        syncPub.publish(syncUpdate);

        if(progressDistance < tolerance)
        {
          currentWaypoint++;
          ROS_INFO("reached new waypoint! have now reached: \t%d/%d\n", currentWaypoint, generatedWaypointNumber);
        }
      }
    }
    else
    {
      waypointMissionUpdateCounter++;
    }
  }
}

void posFeedbackCallback(const geometry_msgs::Transform::ConstPtr& msg)
{
  ex_pos = *msg; // alllllways load the message

  if (!positionMissionFlag)
  {
    return; // peace out if not running mission
  }

  // do loopin' stuff to get close to waypoints
}

bool setupWaypointMission(int responseTimeout)
{
  ros::spinOnce();

  // ROS_INFO();
  while(!ros::param::has("/robot_list/robot_"+to_string(droneId)+"/numWaypoints"))
  {
    ROS_INFO("waiting for waypoints to load...");
  }

  // Waypoint Mission : Initialization
  dji_osdk_ros::MissionWaypointTask waypointTask;
  setWaypointInitDefaults(waypointTask);

  // Waypoint Mission: Create Waypoints
  ROS_INFO("Creating Waypoints..\n");
  generatedWaypts = loadWaypoints();
  // generatedWaypts = createWaypoints();

  // Waypoint Mission: Upload the waypoints
  ROS_INFO("Uploading Waypoints..\n");
  uploadWaypoints(generatedWaypts, responseTimeout, waypointTask);

  // Waypoint Mission: Init mission
  ROS_INFO("Initializing Waypoint Mission..\n");
  if (initWaypointMission(waypointTask).result)
  {
    ROS_INFO("Waypoint upload command sent successfully");
  }
  else
  {
    ROS_WARN("Failed sending waypoint upload command");
    return false;
  }

  return true;
}

// global waypoint mission

bool runWaypointMission()
{

  // Waypoint Mission: Start
  if (missionAction(DJI_MISSION_TYPE::WAYPOINT, MISSION_ACTION::START).result)
  {
    ROS_INFO("Mission start command sent successfully");
  }
  else
  {
    ROS_WARN("Failed sending mission start command");
    return false;
  }

  waypointMissionFlag = true;
  positionMissionFlag = false;
  return true;
}

// relative position mission

bool runPositionMission()
{


  int numPositions;
  ros::param::get("/robot_list/robot_" + to_string(droneId) + "/numPositions", numPositions);
  int currentPositionIndex = 0; // will increment
  ros::Rate updateRate(20); // update period for topics/commands
  positionMissionFlag = true;

  while (positionMissionFlag && ros::ok())
  {
    // print distance to first position
    std::vector<float> currentPos
    {
      static_cast<float>(ex_pos.translation.x),
      static_cast<float>(ex_pos.translation.y),
      static_cast<float>(ex_pos.translation.z)
    };
    std::vector<float> tempPos;
    ros::param::get("/robot_list/robot_" + to_string(droneId) + "/positions/position_" + to_string(currentPositionIndex), tempPos);
    // pull out x,y,z
    std::vector<float> targetPos
    {
      tempPos[1], // 0th index is the index of position, facepalm
      tempPos[2],
      tempPos[3]
    };
    // float targetYaw = tempPos[4]; // TODO: need yaw measurement for this to work1!!!!1!
    // float currentYaw = ex_pos.rotation.
    // map to our coordinate system

    if (!rotationMatrixCalibratedFlag) // could also do this before loop?
    {
      // run rotation matrix calc
      calibratePositionReference();
      rotationMatrixCalibratedFlag = true;
    }

    // convert both target and current position, and move between
    std::vector<float> targetPosMapped = applyRotationMatrixScaling(targetPos);
    std::vector<float> currentPosMapped = applyRotationMatrixScaling(currentPos);
    // map yaw command (in deg) to rotation offset (in rads)
    // float targetYawMapped = applyRotationAngle(targetYaw); // this assumes much, like that we're rotating about the pos z axis

    std::vector<float> moveCommand
    {
      targetPosMapped[0] - currentPosMapped[0],
      targetPosMapped[1] - currentPosMapped[1],
      targetPosMapped[2] - currentPosMapped[2]
    };

    // go to position
    relativePosition(moveCommand[0], moveCommand[1], moveCommand[2], 0.0);

    ros::spinOnce();

    // check if we're close to position
    float checkDist = eucDistance(currentPos, targetPos);
    if (checkDist <= 0.2) // close ish
    {
      currentPositionIndex += 1; // go to next position
    }

    if (currentPositionIndex >= numPositions)
    {
      // leave
      positionMissionFlag = false;
    }
    // ROS_INFO("distance to first position: %f", firstPosDist);


    updateRate.sleep();
  }
  return true;
}

void setWaypointDefaults(WayPointSettings* wp)
{
  wp->damping         = 0.25;
  wp->yaw             = 0.0;
  wp->gimbalPitch     = 0;
  wp->turnMode        = 0;
  wp->hasAction       = 0;
  wp->actionTimeLimit = 100;
  wp->actionNumber    = 0;
  wp->actionRepeat    = 0;
  for (int i = 0; i < 16; ++i)
  {
    wp->commandList[i]      = 0;
    wp->commandParameter[i] = 0;
  }
}

void setWaypointInitDefaults(dji_osdk_ros::MissionWaypointTask& waypointTask)
{
  waypointTask.velocity_range     = 2;
  waypointTask.idle_velocity      = baseSpeedParam;
  waypointTask.action_on_finish   = dji_osdk_ros::MissionWaypointTask::FINISH_NO_ACTION;
  waypointTask.mission_exec_times = 1;
  waypointTask.yaw_mode		  = dji_osdk_ros::MissionWaypointTask::YAW_MODE_RC;
  waypointTask.trace_mode	  = dji_osdk_ros::MissionWaypointTask::TRACE_COORDINATED;
  waypointTask.action_on_rc_lost  = dji_osdk_ros::MissionWaypointTask::ACTION_AUTO;
  waypointTask.gimbal_pitch_mode  = dji_osdk_ros::MissionWaypointTask::GIMBAL_PITCH_FREE;
}

std::vector<DJI::OSDK::WayPointSettings> loadWaypoints()
{

  std::vector<double> wpt; // single waypoint
  // Let's create a vector to store our waypoints in.
  std::vector<DJI::OSDK::WayPointSettings> wp_list;
  int numPts;
  ros::param::get("/robot_list/robot_"+to_string(droneId)+"/numWaypoints", numPts);
  for (int i = 0; i < numPts; i++) {
    ros::param::get("/robot_list/robot_"+to_string(droneId)+"/waypoints/waypoint_" + to_string(i), wpt);
    // ROS_INFO("Waypoint read as (Ind, Lat, Lon, Alt, Yaw): %f %f %f %f %d\n ", wpt[0], wpt[1], wpt[2], wpt[3], wpt[4]);
    WayPointSettings wp;
    setWaypointDefaults(&wp);
    wp.index     = i;
    wp.latitude = wpt[1];
    wp.longitude = wpt[2];
    wp.altitude = wpt[3];
    wp.yaw = wpt[4];
    wp_list.push_back(wp);
  }
  return wp_list;

}

void uploadWaypoints(std::vector<DJI::OSDK::WayPointSettings>& wp_list, int responseTimeout, dji_osdk_ros::MissionWaypointTask& waypointTask)
{
  dji_osdk_ros::MissionWaypoint waypoint;
  double target_yaw = 0.0;
  // for (std::vector<WayPointSettings>::iterator wp = wp_list.begin();
  //      wp != wp_list.end(); ++wp)
  // {
  int numPts;
  DJI::OSDK::WayPointSettings wp;
  ros::param::get("/robot_list/robot_"+to_string(droneId)+"/numWaypoints", numPts);
  for (int i = 0; i < numPts; i++) {

    wp = wp_list[i];
    // target_yaw += 15.0;
    waypoint.latitude            = wp.latitude;
    waypoint.longitude           = wp.longitude;
    waypoint.altitude            = wp.altitude;
    waypoint.damping_distance    = wp.damping;
    waypoint.target_yaw          = wp.yaw;
    waypoint.target_gimbal_pitch = 0;
    waypoint.turn_mode           = wp.turnMode;
    waypoint.has_action          = 0;
    waypointTask.mission_waypoint.push_back(waypoint);
    ROS_INFO("Waypoint created at (Lat, Lon, Alt, Dmp, Yaw): %f %f %f %f %d\n ", waypoint.latitude, waypoint.longitude, waypoint.altitude, waypoint.damping_distance, waypoint.target_yaw);
  }
}

ServiceAck initWaypointMission(dji_osdk_ros::MissionWaypointTask& waypointTask)
{
  dji_osdk_ros::MissionWpUpload missionWpUpload;
  missionWpUpload.request.waypoint_task = waypointTask;
  waypoint_upload_client.call(missionWpUpload);
  if (!missionWpUpload.response.result)
  {
    ROS_WARN("ack.info: set = %i id = %i", missionWpUpload.response.cmd_set,
    missionWpUpload.response.cmd_id);
    ROS_WARN("ack.data: %i", missionWpUpload.response.ack_data);
  }
  return ServiceAck(
    missionWpUpload.response.result, missionWpUpload.response.cmd_set,
    missionWpUpload.response.cmd_id, missionWpUpload.response.ack_data);
}

ServiceAck missionAction(DJI::OSDK::DJI_MISSION_TYPE type,
  DJI::OSDK::MISSION_ACTION   action)
{
  dji_osdk_ros::MissionWpAction missionWpAction;
  dji_osdk_ros::MissionHpAction missionHpAction;
  missionWpAction.request.action = action;
  waypoint_action_client.call(missionWpAction);
  if (!missionWpAction.response.result)
  {
    ROS_WARN("ack.info: set = %i id = %i", missionWpAction.response.cmd_set,
    missionWpAction.response.cmd_id);
    ROS_WARN("ack.data: %i", missionWpAction.response.ack_data);
  }
  return { static_cast<bool>(missionWpAction.response.result),
    missionWpAction.response.cmd_set,
    missionWpAction.response.cmd_id,
    missionWpAction.response.ack_data };
  // switch (type)
  // {
  //   case DJI::OSDK::WAYPOINT:
  // }
}

bool takeoff()
{
  dji_osdk_ros::FlightTaskControl flightTaskControl;
  flightTaskControl.request.task = dji_osdk_ros::FlightTaskControl::Request::TASK_TAKEOFF;
  flight_control_client.call(flightTaskControl);

  return flightTaskControl.response.result;
}

bool land()
{
  dji_osdk_ros::FlightTaskControl flightTaskControl;
  flightTaskControl.request.task = dji_osdk_ros::FlightTaskControl::Request::TASK_FORCE_LANDING;
  flight_control_client.call(flightTaskControl);

  return flightTaskControl.response.result;
}

float applyRotationAngle(float targetYaw)
{
  // make unit vector out of yaw, rotate, dot with pos x, get angle
  const float sinA = sinf( (M_PI/180.0)*targetYaw );
  const float cosA = cosf( (M_PI/180.0)*targetYaw );
  std::vector<float> targetDir = {cosA, sinA, 0.0};
  std::vector<float> rotatedDir = MatrixMultiply(poseRotationMatrix, targetDir);
  // convert back into angle
  std::vector<float> referenceAxis = {1.0, 0.0, 0.0};
  float rotatedYaw = dotProduct(referenceAxis, rotatedDir);
  return (180.0/M_PI)*acosf(rotatedYaw);
}

bool relativePosition(float x, float y, float z, float yaw)
{
  // convert relative position to velocity command
  std::vector<float> posVec {x, y, z};
  float totalDist = Magnitude3(posVec);
  // ROS_INFO("uncapped speed: %f", speedScalarParam*pow(totalDist,2) + minSpeedParam);
  // ROS_INFO("base speed: %f", baseSpeedParam);
  // ROS_INFO("min speed: %f", minSpeedParam);
  // ROS_INFO("speed scalar: %f", speedScalarParam);

  float speed = min(speedScalarParam*pow(totalDist,2) + minSpeedParam, baseSpeedParam);

  float totalTimeMs = (totalDist/speed)*1000.0; // convert from s to ms, aim just short of position
  std::vector<float> velDir = normalizeVector(posVec);
  std::vector<float> velVec = {velDir[0]*speed, velDir[1]*speed, velDir[2]*speed};

  // ROS_INFO("velocity vector: (%f, %f, %f)", velVec[0], velVec[1], velVec[2]);
  // ROS_INFO("speed value: %f", speed);
  //
  // ROS_INFO("distance: %f", totalDist);
  // ROS_INFO("time: %f", totalTimeMs);

  dji_osdk_ros::FlightTaskControl flightTaskControl;
  dji_osdk_ros::JoystickParams joystickParams;
  joystickParams.x = velVec[0];
  joystickParams.y = velVec[1];
  joystickParams.z = velVec[2];
  joystickParams.yaw = yaw; // for now until i have yaw measurement up
  flightTaskControl.request.task = dji_osdk_ros::FlightTaskControl::Request::TASK_VELOCITY_AND_YAWRATE_CONTROL;
  flightTaskControl.request.joystickCommand = joystickParams;
  flightTaskControl.request.velocityControlTimeMs = totalTimeMs;
  flightTaskControl.request.posThresholdInM = 0.1;
  flightTaskControl.request.yawThresholdInDeg = 1;

  flight_control_client.call(flightTaskControl); // actually waits for drone to "reach" position

  return flightTaskControl.response.result;
}

bool stopMotion()
{
  if (waypointMissionFlag)
  {
    dji_osdk_ros::MissionWpAction waypointAction;
    waypointAction.request.action = dji_osdk_ros::MissionWpAction::Request::ACTION_STOP; // stahhp
    waypoint_action_client.call(waypointAction);
    waypointMissionFlag = false;
  }

  if (positionMissionFlag)
  {
    positionMissionFlag = false;
  }
  return true;
}

bool goHome()
{
  // TODO fix this pls
  return true;
}

void calibratePositionReference()
{
  // call move and check what pos ref says
  ros::spinOnce(); // let callbacks update
  geometry_msgs::Transform preT = ex_pos;
  relativePosition(0.5, 0.0, 0.0, 0.0);
  ros::spinOnce(); // let callbacks update
  geometry_msgs::Transform postT = ex_pos;

  std::vector<float> requestedPosition {0.5, 0.0, 0.0}; // what we asked for
  std::vector<float> reversePosition {-0.5, 0.0, 0.0}; // what we asked for
  std::vector<float> actualPosition // what we got
  {
    static_cast<float>(postT.translation.x) - static_cast<float>(preT.translation.x),
    static_cast<float>(postT.translation.y) - static_cast<float>(preT.translation.y),
    static_cast<float>(postT.translation.z) - static_cast<float>(preT.translation.z),
  };

  // calculate rotation matrix
  poseRotationMatrix = rotateAlign(requestedPosition, actualPosition);

  // now get scaling difference between coordinate frames
  float requestedPositionMag = Magnitude3(requestedPosition);
  float actualPositionMag = Magnitude3(actualPosition);

  poseMagnitudeScaler = (requestedPositionMag/actualPositionMag)*0.8; // aim just short of actual scalar, better to under than over shoot

  ROS_INFO("pose magnitude scalar %f", poseMagnitudeScaler);

  std::vector<float> checkPosition = applyRotationMatrixScaling(requestedPosition);
  relativePosition(reversePosition[0], reversePosition[1], reversePosition[2], 0.0); // go to requested position in corrected frame
  ros::spinOnce();

  // write rotation and scale to parameters
  // for (int i = 0; i < poseRotationMatrix.size(); i++) {
  //   for (int j = 0; j < poseRotationMatrix[0].size(); j++) {
  //     ros::param::set("poseRotationMatrix/val_" + to_string(i) + to_string(j), poseRotationMatrix[i][j]);
  //   }
  // }
  // ros::param::set("poseMagnitudeScaler", poseMagnitudeScaler);
}

std::vector<float> applyRotationMatrixScaling(std::vector<float> inputVec)
{
  std::vector<float> rotatedVec = MatrixMultiply(poseRotationMatrix, inputVec);
  std::vector<float> outputVec
  {
    rotatedVec[0]*poseMagnitudeScaler,
    rotatedVec[1]*poseMagnitudeScaler,
    rotatedVec[2]*poseMagnitudeScaler
  };
  return outputVec;
}

// bool add(beginner_tutorials::AddTwoInts::Request &req, beginner_tutorials::AddTwoInts::Response &res)
bool overwatchFunction(dji_osdk_ros::Overwatch::Request &req, dji_osdk_ros::Overwatch::Response &res)
{

  // always request control?
  dji_osdk_ros::ObtainControlAuthority obtainCtrlAuthority;
  obtainCtrlAuthority.request.enable_obtain = true;
  obtain_ctrl_authority_client.call(obtainCtrlAuthority);

  // switch based on requested action
  switch (req.action) {
    case 1:
    {
      ROS_INFO("received overwatch: start waypoint");
      // start waypoint
      int responseTimeout = 1;
      setupWaypointMission(responseTimeout); // for now!
      runWaypointMission();
      break;
    }
    case 2:
    {
      ROS_INFO("received overwatch: start position");
      // start position
      runPositionMission();
      break;
    }
    case 3:
    {
      ROS_INFO("received overwatch: stop all motion");
      // stop all motion
      stopMotion();
      break;
    }
    case 4:
    {
      ROS_INFO("received overwatch: land");
      // land
      land();
      break;
    }
    case 5:
    {
      ROS_INFO("received overwatch: go home");
      // go home
      goHome();
      break;
    }
    case 6:
    {
      ROS_INFO("received overwatch: takeoff");
      // go home
      takeoff();
      break;
    }
    case 7:
    {
      ROS_INFO("received overwatch: calibrate pos ref");
      // go home
      calibratePositionReference();
      break;
    }
    default:
      break;
  }

  res.result = true;
  return res.result;
}

int main(int argc, char** argv)
{

  ros::init(argc, argv, "mission_node");
  ros::NodeHandle nh;

  ros::param::get("base_speed", baseSpeedParam);
  ros::param::get("min_speed", minSpeedParam);
  ros::param::get("speed_scalar", speedScalarParam);
  ros::param::get("drone_id", droneId);
  // ROS_INFO("base speed set to %f", baseSpeedParam);

  // ROS stuff
  // obtain control authority?
  obtain_ctrl_authority_client = nh.serviceClient<dji_osdk_ros::ObtainControlAuthority>(
    "obtain_release_control_authority");

  waypoint_upload_client = nh.serviceClient<dji_osdk_ros::MissionWpUpload>(
    "dji_osdk_ros/mission_waypoint_upload");

  waypoint_action_client = nh.serviceClient<dji_osdk_ros::MissionWpAction>(
    "dji_osdk_ros/mission_waypoint_action");

  flight_control_client = nh.serviceClient<dji_osdk_ros::FlightTaskControl>(
    "flight_task_control");

  syncPub = nh.advertise<std_msgs::Float32>("sync_signal", 1);
  gps_pos_subscriber = nh.subscribe<sensor_msgs::NavSatFix>("dji_osdk_ros/gps_position", 1, &gpsPosCallback);
  pos_feedback_subscriber = nh.subscribe<geometry_msgs::Transform>("pose_feedback", 1, &posFeedbackCallback);

  ros::ServiceServer overwatchService = nh.advertiseService("overwatch", overwatchFunction);

  // wait for services to be up
  obtain_ctrl_authority_client.waitForExistence();
  waypoint_upload_client.waitForExistence();
  waypoint_action_client.waitForExistence();
  flight_control_client.waitForExistence();
  // wait for waypoints to be up as well
  ROS_INFO("bus driver ready to go!");

  ros::spin();

  return 0;
}
